'''This contains both the LSSTAuthManager class and a handful of
functions that are basically slightly-tailored versions of set
manipulation, designed for doing things with user group membership.
'''

import json
from asgiref.sync import async_to_sync
from eliot import start_action
from .. import LoggableChild
from ..utils import sanitize_dict


class LSSTAuthManager(LoggableChild):
    '''Class to hold LSST-specific authentication/authorization details
    and methods.

    Most of this was formerly held in the JupyterHub config as classes defined
    in '10-authenticator.py'.
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.authenticator = self.parent.authenticator
        self.uid = None
        self.group_map = {}
        self.pod_env = {}
        # Do not set self.auth_state.  It gets incorrectly cached.

    def get_group_string(self):
        '''Convenience function for retrieving the group name-to-uid mapping
        list as a string suitable for passing to the spawned pod.
        '''
        with start_action(action_type="get_group_string"):
            return ','.join(["{}:{}".format(x, self.group_map[x])
                             for x in self.group_map])

    def get_pod_env(self):
        '''Return the authenticator-specific fields.
        '''
        with start_action(action_type="get_pod_env"):
            return self.pod_env

    def parse_auth_state(self):
        '''Take the auth_state from parent.spawner and extract:
            * UID
            * Group/gid mappings
            * Possibly-authenticator-specific fields for pod environment

        and then store them in pod_env for the pod to pick up at spawn
         time. Our groups will be set if we authenticated, but if we got to
         the spawner page via a user that was already authenticated in the
         session DB (that is, you killed your pod and went away, but didn't
         log out, and then came back later while your session was still valid
         but the Hub had restarted), the authenticate() method in the spawner
         won't be run (since your user is still valid) but the fields won't
         be set (because the LSST managers are not persisted).  Hence the
         group mapping re-check, because to get around exactly this problem,
         each authenticator stores the group string in auth_state.
        '''
        with start_action(action_type="parse_auth_state"):
            self.log.debug("Parsing authentication state.")
            pod_env = {}
            # Force refresh if we have never set uid
            user = self.parent.spawner.user
            ast = async_to_sync(user.get_auth_state)()
            self.uid = ast["uid"]
            if not self.uid:
                raise RuntimeError("Cannot determine user UID for pod spawn!")
            claims = ast["claims"]
            token = ast["access_token"]
            email = claims.get("email") or ''
            pod_env['ACCESS_TOKEN'] = token
            pod_env['GITHUB_EMAIL'] = email
            self.group_map = ast["group_map"]
            if not self.group_map:
                raise RuntimeError("Cannot determine user GIDs for pod spawn!")
            groupstr = self.get_group_string()
            pod_env['EXTERNAL_UID'] = str(self.uid)
            pod_env['EXTERNAL_GROUPS'] = groupstr
            self.pod_env = pod_env

    def dump(self):
        '''Return dict of contents for pretty-printing.
        '''
        ast = async_to_sync(self.parent.spawner.user.get_auth_state)()
        pd = {"parent": str(self.parent),
              "uid": self.uid,
              "group_map": self.group_map,
              "auth_state": sanitize_dict(ast, ['access_token']),
              "pod_env": sanitize_dict(self.pod_env,
                                       ['ACCESS_TOKEN',
                                        'GITHUB_ACCESS_TOKEN'])}

        return pd

    def toJSON(self):
        return json.dumps(self.dump())



# Inspired by, and based on, Adam Tilghman's Multi-Namespace work in
#  https://github.com/jupyterhub/kubespawner/pull/218
import threading
import time
from eliot import start_action
from kubernetes import watch
from kubespawner.reflector import NamespacedResourceReflector
from traitlets import Bool
# This is kubernetes client implementation specific, but we need to know
# whether it was a network or watch timeout.
from urllib3.exceptions import ReadTimeoutError


class MultiNamespaceResourceReflector(NamespacedResourceReflector):
    list_method_omit_namespace = Bool(
        False,
        help="""
        If True, our calls to API `list_method_name` will omit a `namespace`
        argument.  Necessary for non-namespaced methods such as
        `list_pod_for_all_namespaces`
        """
    )

    def _start_watching_pods(self, replace=False):
        """Start the pod reflector

        If replace=False and the pod reflector is already running,
        do nothing.

        If replace=True, a running pod reflector will be stopped
        and a new one started (for recovering from possible errors).
        """
        return self._start_reflector("pods", MultiNamespacePodReflector,
                                     replace=replace)

    def _create_resource_key(self, resource):
        """Maps a Kubernetes resource object onto a hashable Dict key;
        subclass may override if `resource.metadata.name` is not
        unique (e.g. pods across multiple namespaces)
        """
        return resource.metadata.name

    def _list_and_update(self):
        """
        Update current list of resources by doing a full fetch.
        Overwrites all current resource info.
        """
        # Way too spammy to log with eliot.
        list_args = {
            'label_selector': self.label_selector,
            'field_selector': self.field_selector,
            '_request_timeout': self.request_timeout
        }

        if not self.list_method_omit_namespace:
            list_args['namespace'] = self.namespace

        initial_resources = getattr(
            self.api, self.list_method_name)(**list_args)

        # This is an atomic operation on the dictionary!
        self.resources = {self._create_resource_key(
            p): p for p in initial_resources.items}
        # return the resource version so we can hook up a watch
        return initial_resources.metadata.resource_version

    def _watch_and_update(self):
        """
        Keeps the current list of resources up-to-date

        This method is to be run not on the main thread!

        We first fetch the list of current resources, and store that. Then we
        register to be notified of changes to those resources, and keep our
        local store up-to-date based on these notifications.
        We also perform exponential backoff, giving up after we hit 32s
        wait time. This should protect against network connections dropping
        and intermittent unavailability of the api-server. Every time we
        recover from an exception we also do a full fetch, to pick up
        changes that might've been missed in the time we were not doing
        a watch.

        Note that we're playing a bit with fire here, by updating a dictionary
        in this thread while it is probably being read in another thread
        without using locks! However, dictionary access itself is atomic,
        and as long as we don't try to mutate them (do a 'fetch / modify /
        update' cycle on them), we should be ok!
        """
        with start_action(action_type="_watch_and_update"):
            selectors = []
            if self.label_selector:
                selectors.append("label selector=%r" % self.label_selector)
            if self.field_selector:
                selectors.append("field selector=%r" % self.field_selector)
            cur_delay = 0.1
            while True:
                start = time.monotonic()
                w = watch.Watch()
                try:
                    resource_version = self._list_and_update()
                    if not self.first_load_future.done():
                        # signal that we've loaded our initial data
                        self.first_load_future.set_result(None)
                        watch_args = {
                            'label_selector': self.label_selector,
                            'field_selector': self.field_selector,
                            'resource_version': resource_version,
                        }
                    if not self.list_method_omit_namespace:
                        watch_args['namespace'] = self.namespace
                    if self.request_timeout:
                        # set network receive timeout
                        watch_args['_request_timeout'] = self.request_timeout
                    if self.timeout_seconds:
                        # set watch timeout
                        watch_args['timeout_seconds'] = self.timeout_seconds
                    # in case of timeout_seconds, the w.stream just exits
                    #  (no exception thrown)
                    #     -> we stop the watcher and start a new one
                    for ev in w.stream(
                            getattr(self.api, self.list_method_name),
                            **watch_args
                    ):
                        cur_delay = 0.1
                        resource = ev['object']
                        if ev['type'] == 'DELETED':
                            # This is an atomic delete operation on the
                            # dictionary!
                            self.resources.pop(
                                self._create_resource_key(resource), None)
                        else:
                            # This is an atomic operation on the dictionary!
                            self.resources[self._create_resource_key(
                                resource)] = resource
                            if self._stop_event.is_set():
                                break
                    watch_duration = time.monotonic() - start
                    if watch_duration >= self.restart_seconds:
                        break
                except ReadTimeoutError:
                    # network read time out, just continue and restart
                    # the watch; this could be due to a network problem
                    # or just low activity
                    continue
                except Exception:
                    cur_delay = cur_delay * 2
                    if cur_delay > 30:
                        self.log.exception(
                            "Watching resources never recovered, giving up")
                        if self.on_failure:
                            self.on_failure()
                        return
                    self.log.exception(
                        "Error when watching resources, retrying in" +
                        " %ss" % cur_delay)
                    time.sleep(cur_delay)
                    continue
                else:
                    # no events on watch, reconnect
                    pass
                finally:
                    w.stop()
                    if self._stop_event.is_set():
                        break

    def start(self):
        """
        Start the reflection process!

        We'll do a blocking read of all resources first, so that we don't
        race with any operations that are checking the state of the pod
        store - such as polls. This should be called only once at the
        start of program initialization (when the singleton is being created),
        and not afterwards!
        """
        if hasattr(self, 'watch_thread'):
            raise ValueError('Thread watching for resources is already running'
                             )

        self._list_and_update()
        self.watch_thread = threading.Thread(target=self._watch_and_update)
        # If the watch_thread is only thread left alive, exit app
        self.watch_thread.daemon = True
        self.watch_thread.start()


class MultiNamespaceEventReflector(MultiNamespaceResourceReflector):
    kind = 'events'

    list_method_name = 'list_namespaced_event'

    @property
    def events(self):
        return sorted(
            self.resources.values(),
            key=lambda event: event.last_timestamp or event.event_time,
        )


class MultiNamespacePodReflector(MultiNamespaceResourceReflector):
    kind = 'pods'
    # FUTURE: These labels are the selection labels for the PodReflector. We
    # might want to support multiple deployments in the same namespace, so we
    # would need to select based on additional labels such as `app` and
    # `release`.
    labels = {
        'component': 'singleuser-server',
    }
    list_method_name = 'list_pod_for_all_namespaces'
    list_method_omit_namespace = True

    @property
    def pods(self):
        return self.resources

    def _create_resource_key(self, resource):
        return (resource.metadata.namespace, resource.metadata.name)



'''
Shared utility functions.
'''

import hashlib
import inspect
import logging
import os
import random
import requests

from collections import defaultdict
from eliot.stdlib import EliotHandler


def rreplace(s, old, new, occurrence):
    '''Convenience function from:
    https://stackoverflow.com/questions/2556108/\
    rreplace-how-to-replace-the-last-occurrence-of-an-expression-in-a-string
    '''
    li = s.rsplit(old, occurrence)
    return new.join(li)


def sanitize_dict(input_dict, sensitive_fields):
    '''Remove sensitive content.  Useful for logging.
    '''
    retval = {}
    if not input_dict:
        return retval
    retval.update(input_dict)
    for field in sensitive_fields:
        if retval.get(field):
            retval[field] = "[redacted]"
    return retval


def get_execution_namespace():
    '''Return Kubernetes namespace of this container.
    '''
    ns_path = '/var/run/secrets/kubernetes.io/serviceaccount/namespace'
    if os.path.exists(ns_path):
        with open(ns_path) as f:
            return f.read().strip()
    return None


def make_logger(name=None, level=None):
    '''Create a logger with LSST-appropriate characteristics.
    '''
    if name is None:
        # Get name of caller's class.
        #  From https://stackoverflow.com/questions/17065086/
        frame = inspect.stack()[1][0]
        name = _get_classname_from_frame(frame)
    logger = logging.getLogger(name)
    logger.propagate = False
    if level is None:
        level = logging.getLogger().getEffectiveLevel()
    logger.setLevel(level)
    logger.handlers = [EliotHandler()]
    logger.info("Created logger object for class '{}'.".format(name))
    return logger


def _get_classname_from_frame(fr):
    args, _, _, value_dict = inspect.getargvalues(fr)
    # we check the first parameter for the frame function is
    # named 'self'
    if len(args) and args[0] == 'self':
        # in that case, 'self' will be referenced in value_dict
        instance = value_dict.get('self', None)
        if instance:
            # return its classname
            cl = getattr(instance, '__class__', None)
            if cl:
                return "{}.{}".format(cl.__module__, cl.__name__)
    # If it wasn't a class....
    return '<unknown>'


def str_bool(s):
    '''Make a sane guess for whether a value represents true or false.
    Intended for strings, mostly in the context of environment variables,
    but if you pass it something that's not a string that is falsy, like
    an empty list, it will cheerfully return False.
    '''
    if not s:
        return False
    if type(s) != str:
        # It's not a string and it's not falsy, soooo....
        return True
    s = s.lower()
    if s in ['false', '0', 'no', 'n']:
        return False
    return True


def str_true(v):
    '''The string representation of a true value will be 'TRUE'.  False will
    be the empty string.
    '''
    if v:
        return 'TRUE'
    else:
        return ''


def listify(item, delimiter=','):
    '''Used for taking character (usually comma)-separated string lists
    and returning an actual list, or the empty list.
    Useful for environment parsing.

    Sure, you could pass it integer zero and get [] back.  Don't.
    '''
    if not item:
        return []
    if type(item) is str:
        item = item.split(delimiter)
    if type(item) is not list:
        raise TypeError("'listify' must take None, str, or list!")
    return item


def floatify(item, default=0.0):
    '''Another environment-parser: the empty string should be treated as
    None, and return the default, rather than the empty string (which
    does not become an integer).  Default can be either a float or string
    that float() works on.  Note that numeric zero (or string '0') returns
    0.0, not the default.  This is intentional.
    '''
    if item is None:
        return default
    if item == '':
        return default
    return float(item)


def intify(item, default=0):
    '''floatify, but for ints.
    '''
    return int(floatify(item, default))


def list_duplicates(seq):
    '''List duplicate items from a sequence.
    '''
    # https://stackoverflow.com/questions/5419204
    tally = defaultdict(list)
    for i, item in enumerate(seq):
        tally[item].append(i)
    return ((key, locs) for key, locs in tally.items()
            if len(locs) > 1)


def list_digest(inp_list):
    '''Return a digest to uniquely identify a list.
    '''
    if type(inp_list) is not list:
        raise TypeError("list_digest only works on lists!")
    if not inp_list:
        raise ValueError("input must be a non-empty list!")
    # If we can rely on python >= 3.8, shlex.join is better
    return hashlib.sha256(' '.join(inp_list).encode('utf-8')).hexdigest()


def get_access_token(tokenfile=None):
    '''Determine the access token from the mounted secret or environment.
    '''
    tok = None
    hdir = os.environ.get('HOME', None)
    if hdir:
        if not tokenfile:
            # FIXME we should make this instance-dependent
            tokfile = hdir + "/.access_token"
        try:
            with open(tokfile, 'r') as f:
                tok = f.read().replace('\n', '')
        except Exception as exc:
            log = make_logger()
            log.warn("Could not read tokenfile '{}': {}".format(tokfile, exc))
    if not tok:
        tok = os.environ.get('ACCESS_TOKEN', None)
    return tok


def parse_access_token(endpoint=None, tokenfile=None, token=None, timeout=15):
    '''Rely on gafaelfawr to validate and parse the access token.
    '''
    if not token:
        token = get_access_token(tokenfile=tokenfile)
    if not token:
        raise RuntimeError("Cannot determine access token!")
    # Endpoint is constant in an ArgoCD-deployed cluster
    if not endpoint:
        endpoint = "http://gafaelfawr-service.gafaelfawr:8080/auth/analyze"
    resp = requests.post(endpoint, data={'token': token}, timeout=timeout)
    rj = resp.json()
    p_resp = rj["token"]
    if not p_resp["valid"]:
        raise RuntimeError("Access token invalid: '{}'!".format(str(resp)))
    # Force to lowercase username (should no longer be necessary)
    p_tok = p_resp["data"]
    uname = p_tok["uid"]
    p_tok["uid"] = uname.lower()
    return p_tok


def assemble_gids(groupinfo):
    '''Take the group data (in the format of a CILogon "isMemberOf" claim)
    and return the string to be used for provisioning the user.

    Simply ignore names without corresponding ids.
    '''
    gidlist = ["{}:{}".format(x['name'], x['id'])
               for x in groupinfo if 'id' in x]
    return ','.join(gidlist)


def get_fake_gid():
    '''Use if we have strict_ldap_groups off, to assign GIDs to names
    with no matching Unix GID.  Since these will not appear as filesystem
    groups, being consistent with them isn't important.  We just need
    to make their GIDs something likely to not match anything real.

    There is a chance of collision, but it doesn't really matter.

    We do need to keep the no-GID groups around, though, because we might
    be using them to make options form or quota decisions (if we know we
    don't, we should turn on strict_ldap_groups).
    '''
    grpbase = 3E7
    grprange = 1E7
    igrp = random.randint(grpbase, (grpbase + grprange))
    return igrp
